import contextlib
import sys

from boa.env import Env, enable_pyevm_verbose_logging
from boa.interpret import BoaError, load, load_partial, loads, loads_partial

# turn off tracebacks if we are in repl
# https://stackoverflow.com/a/64523765
if hasattr(sys, "ps1"):
    pass
    # sys.tracebacklimit = 0

env = Env.get_singleton()


@contextlib.contextmanager
def reverts():
    try:
        yield
        raise Exception("Did not revert")
    except BoaError:
        pass


def eval(code):
    return loads("").eval(code)



from eth_utils import to_checksum_address
from vyper.codegen.types.types import (
    ByteArrayType,
    DArrayType,
    SArrayType,
    StringType,
    is_base_type,
    is_bytes_m_type,
    is_integer_type,
)
from vyper.utils import unsigned_to_signed


def ceil32(n):
    return floor32(n + 31)


def floor32(n):
    return n & ~31


# wrap storage in something which looks like memory
class ByteAddressableStorage:
    def __init__(self, db, address, key):
        self.db = db
        self.address = address
        self.key = key

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            ret = b""
            start = subscript.start or 0
            stop = subscript.stop
            i = self.key + start // 32
            while i < self.key + ceil32(stop) // 32:
                ret += self.db.get_storage(self.address, i).to_bytes(32, "big")
                i += 1

            start -= floor32(start)
            stop -= floor32(start)
            return memoryview(ret[start:stop])
        else:
            raise Exception("Must slice {self}")


def decode_vyper_object(mem, typ):
    if is_bytes_m_type(typ):
        # TODO tag return value like `vyper_object` does
        return mem[: typ._bytes_info.m].tobytes()
    if is_base_type(typ, "address"):
        return to_checksum_address(mem[12:32].tobytes())
    if is_base_type(typ, "bool"):
        return bool.from_bytes(mem[31:32], "big")
    if is_integer_type(typ):
        ret = int.from_bytes(mem[:32], "big")
        if typ._int_info.is_signed:
            return unsigned_to_signed(ret, 256)
        return ret
    if isinstance(typ, ByteArrayType):
        length = int.from_bytes(mem[:32], "big")
        return mem[32 : 32 + length].tobytes()
    if isinstance(typ, StringType):
        length = int.from_bytes(mem[:32], "big")
        return mem[32 : 32 + length].tobytes().decode("utf-8")
    if isinstance(typ, SArrayType):
        length = typ.count
        n = typ.subtype.memory_bytes_required
        return [
            decode_vyper_object(mem[i * n : i * n + n], typ.subtype)
            for i in range(length)
        ]
    if isinstance(typ, DArrayType):
        length = int.from_bytes(mem[:32], "big")
        n = typ.subtype.memory_bytes_required
        ofst = 32
        ret = []
        for _ in range(length):
            ret.append(decode_vyper_object(mem[ofst : ofst + n], typ.subtype))
            ofst += n
        return ret

    return "unimplemented decoder for `{typ}`"



from typing import Optional, Union

import vyper
from vyper.compiler.phases import CompilerData

from boa.contract import BoaError, VyperBlueprint, VyperContract, VyperDeployer
from boa.util.disk_cache import DiskCache

_Contract = Union[VyperContract, VyperBlueprint]


_disk_cache = None


def set_cache_dir(cache_dir="~/.cache/titanoboa"):
    global _disk_cache
    if cache_dir is None:
        _disk_cache = None
        return
    compiler_version = f"{vyper.__version__}.{vyper.__commit__}"
    _disk_cache = DiskCache(cache_dir, compiler_version)


def compiler_data(source_code: str, contract_name: str) -> CompilerData:
    global _disk_cache

    if _disk_cache is None:
        return CompilerData(source_code, contract_name)

    def func():
        ret = CompilerData(source_code, contract_name)
        ret.bytecode_runtime  # force compilation to happen
        return ret

    return _disk_cache.caching_lookup(source_code, func)


def load(filename: str, *args, **kwargs) -> _Contract:  # type: ignore
    with open(filename) as f:
        return loads(f.read(), *args, name=filename, **kwargs)


def loads(source_code, *args, as_blueprint=False, name=None, **kwargs):
    d = loads_partial(source_code, name)
    if as_blueprint:
        return d.deploy_as_blueprint(**kwargs)
    else:
        return d.deploy(*args, **kwargs)


def loads_partial(source_code: str, name: Optional[str] = None) -> VyperDeployer:
    name = name or "VyperContract"  # TODO handle this upstream in CompilerData
    data = compiler_data(source_code, name)
    return VyperDeployer(data)


def load_partial(filename: str) -> VyperDeployer:  # type: ignore
    with open(filename) as f:
        return loads_partial(f.read(), name=filename)


__all__ = ["BoaError"]



# TODO maybe move me to boa/vyper/

import contextlib
import copy
import textwrap
import warnings
from typing import Any

import eth_abi as abi
import vyper
import vyper.ast as vy_ast
import vyper.compiler.output as compiler_output
import vyper.ir.compile_ir as compile_ir
import vyper.semantics.namespace as vy_ns
import vyper.semantics.validation as validation
from eth_utils import to_canonical_address, to_checksum_address
from vyper.ast.signatures.function_signature import FunctionSignature
from vyper.ast.utils import parse_to_ast
from vyper.codegen.core import calculate_type_for_external_return, getpos
from vyper.codegen.function_definitions.common import generate_ir_for_function
from vyper.codegen.ir_node import IRnode
from vyper.codegen.module import parse_external_interfaces
from vyper.codegen.types.types import MappingType, TupleType
from vyper.exceptions import InvalidType, VyperException
from vyper.ir.optimizer import optimize
from vyper.semantics.validation.data_positions import set_data_positions
from vyper.semantics.validation.utils import get_exact_type_from_node
from vyper.utils import abi_method_id, cached_property

from boa.env import AddressT, Env, to_int
from boa.vyper.decoder_utils import ByteAddressableStorage, decode_vyper_object


# build a reverse map from the format we have in pc_pos_map to AST nodes
def ast_map_of(ast_node):
    ast_map = {}
    nodes = [ast_node] + ast_node.get_descendants(reverse=True)
    for node in nodes:
        ast_map[getpos(node)] = node
    return ast_map


# error messages for external calls
EXTERNAL_CALL_ERRORS = ("external call failed", "returndatasize too small")


# id used internally for method id name
_METHOD_ID_VAR = "_calldata_method_id"


class lrudict(dict):
    def __init__(self, n, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n = n

    def __getitem__(self, k):
        val = super().__getitem__(k)
        del self[k]
        super().__setitem__(k, val)
        return val

    def __setitem__(self, k, val):
        if len(self) == self.n:
            del self[next(iter(self))]
        super().__setitem__(k, val)


class VyperDeployer:
    def __init__(self, compiler_data, env=None):
        self.compiler_data = compiler_data

    def deploy(self, *args, **kwargs):
        return VyperContract(self.compiler_data, *args, **kwargs)

    def deploy_as_blueprint(self, *args, **kwargs):
        return VyperBlueprint(self.compiler_data, *args, **kwargs)

    def at(self, address: AddressT) -> "VyperContract":
        address = to_checksum_address(address)
        ret = VyperContract(
            self.compiler_data, override_address=address, skip_init=True
        )
        vm = ret.env.vm
        bytecode = vm.state.get_code(to_canonical_address(address))

        ret._set_bytecode(bytecode)
        return ret


# a few lines of shared code between VyperBlueprint and VyperContract
class _BaseContract:
    def __init__(self, compiler_data, env=None, override_address=None):
        self.compiler_data = compiler_data

        if env is None:
            env = Env.get_singleton()

        self.env = env
        if override_address is None:
            self.address = self.env.generate_address()
        else:
            self.address = override_address


# create a blueprint for use with `create_from_blueprint`.
# uses a ERC5202 preamble, when calling `create_from_blueprint` will
# need to use `code_offset=3`
class VyperBlueprint(_BaseContract):
    def __init__(
        self,
        compiler_data,
        env=None,
        override_address=None,
        blueprint_preamble=b"\xFE\x71\x00",
    ):
        # note slight code duplication with VyperContract ctor,
        # maybe use common base class?
        super().__init__(compiler_data, env, override_address)

        if blueprint_preamble is None:
            blueprint_preamble = b""

        blueprint_bytecode = blueprint_preamble + compiler_data.bytecode

        # the length of the deployed code in bytes
        len_bytes = len(blueprint_bytecode).to_bytes(2, "big")
        deploy_bytecode = b"\x61" + len_bytes + b"\x3d\x81\x60\x0a\x3d\x39\xf3"

        deploy_bytecode += blueprint_bytecode

        self.bytecode = self.env.deploy_code(
            bytecode=deploy_bytecode, deploy_to=self.address
        )


class FrameDetail(dict):
    def __init__(self, fn_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fn_name = fn_name

    def __repr__(self):
        detail = ", ".join(f"{k}={v}" for (k, v) in self.items())
        return f"<{self.fn_name}: {detail}>"


class StackTrace(list):
    def __repr__(self):
        return "\n".join(repr(x) for x in self)


def unwrap_storage_key(sha3_db, k):
    path = []

    def unwrap(k):
        if k in sha3_db:
            preimage = sha3_db[k]
            slot, k = preimage[:32], preimage[32:]

            unwrap(slot)

        path.append(k)

    unwrap(k)
    return path


def setpath(lens, path, val):
    for i, k in enumerate(path):
        if i == len(path) - 1:
            lens[k] = val
        else:
            lens = lens.setdefault(k, {})


class VarModel:
    def __init__(self, contract, slot, typ):
        self.contract = contract
        self.addr = to_canonical_address(self.contract.address)
        self.accountdb = contract.env.vm.state._account_db
        self.slot = slot
        self.typ = typ

    def _decode(self, slot, typ):
        fakemem = ByteAddressableStorage(self.accountdb, self.addr, slot)
        return decode_vyper_object(fakemem, typ)

    def _dealias(self, maybe_address):
        try:
            return self.contract.env.lookup_alias(maybe_address)
        except:  # not found, return the input
            return maybe_address

    def get(self):
        if isinstance(self.typ, MappingType):
            ret = {}
            for k in self.contract.env.sstore_trace.get(self.contract.address, {}):
                path = unwrap_storage_key(self.contract.env.sha3_trace, k)
                if to_int(path[0]) != self.slot:
                    continue

                path = path[1:]

                ty = self.typ
                for i, p in enumerate(path):
                    path[i] = decode_vyper_object(memoryview(p), ty.keytype)
                    ty = ty.valuetype

                val = self._decode(to_int(k), ty)

                if val:
                    # decode aliases
                    path = [self._dealias(p) for p in path]
                    setpath(ret, path, val)
            return ret

        else:
            return self._decode(self.slot, self.typ)


# data structure to represent the storage variables in a contract
class StorageModel:
    def __init__(self, contract):
        compiler_data = contract.compiler_data
        for k, v in compiler_data.global_ctx._globals.items():
            is_storage = not v.is_immutable
            if is_storage:  # check that v
                slot = compiler_data.storage_layout["storage_layout"][k]["slot"]
                setattr(self, k, VarModel(contract, slot, v.typ))


class VyperContract(_BaseContract):
    def __init__(
        self, compiler_data, *args, env=None, override_address=None, skip_init=False
    ):
        super().__init__(compiler_data, env, override_address)

        self.env.register_contract(self.address, self)

        fns = {fn.name: fn for fn in self.global_ctx._function_defs}
        # add all functions from the interface to the contract

        self._ctor = None
        if "__init__" in fns:
            self._ctor = VyperFunction(fns.pop("__init__"), self)

        for fn in fns.values():
            setattr(self, fn.name, VyperFunction(fn, self))

        self._storage = StorageModel(self)

        self._eval_cache = lrudict(0x1000)
        self._source_map = None
        self._computation = None

        if not skip_init:
            self._run_init(*args)

    def _run_init(self, *args):
        encoded_args = b""

        if self._ctor:
            encoded_args = self._ctor._prepare_calldata(*args)
            encoded_args = encoded_args[4:]  # strip method_id

        initcode = self.compiler_data.bytecode + encoded_args
        self.bytecode = self.env.deploy_code(bytecode=initcode, deploy_to=self.address)

    # manually set the runtime bytecode, instead of using deploy
    def _set_bytecode(self, bytecode: bytes) -> None:
        if bytecode[-self.data_section_size :] != self.compiler_data.bytecode_runtime:
            warnings.warn(f"casted bytecode does not match compiled bytecode at {self}")
        self.bytecode = bytecode

    def __str__(self):
        return (
            f"<{self.compiler_data.contract_name} at {to_checksum_address(self.address)}, "
            f"compiled with vyper-{vyper.__version__}+{vyper.__commit__}>"
        )

    @cached_property
    def deployer(self):
        return VyperDeployer(self.compiler_data, env=self.env)

    # is this actually useful?
    def at(self, *args, **kwargs):
        return self.deployer.wrap(*args, **kwargs)

    @cached_property
    def ast_map(self):
        return ast_map_of(self.compiler_data.vyper_module)

    def debug_frame(self, computation=None):
        if computation is None:
            computation = self._computation

        node = self.find_source_of(computation.code)
        if node is None:
            return None

        fn = node.get_ancestor(vy_ast.FunctionDef)

        frame_info = self.compiler_data.function_signatures[fn.name].frame_info

        mem = computation._memory
        frame_detail = FrameDetail(fn.name)
        for k, v in frame_info.frame_vars.items():
            if v.location.name != "memory":
                continue
            ofst = v.pos
            size = v.typ.memory_bytes_required
            frame_detail[k] = decode_vyper_object(mem.read(ofst, size), v.typ)

        return frame_detail

    @property
    def global_ctx(self):
        return self.compiler_data.global_ctx

    @property
    def source_map(self):
        if self._source_map is None:
            _, self._source_map = compile_ir.assembly_to_evm(
                self.compiler_data.assembly_runtime
            )
        return self._source_map

    def find_error_meta(self, code_stream):
        error_map = self.source_map.get("error_map", {})
        for pc in reversed(code_stream._trace):
            if pc in error_map:
                return error_map[pc]
        return None

    def find_source_of(self, code_stream, is_initcode=False):
        pc_map = self.source_map["pc_pos_map"]
        for pc in reversed(code_stream._trace):
            if pc in pc_map and pc_map[pc] in self.ast_map:
                return self.ast_map[pc_map[pc]]

        return None

    def marshal_to_python(self, computation, vyper_typ):
        self._computation = computation  # for further inspection

        if computation.is_error:
            self.handle_error(computation)

        if vyper_typ is None:
            return None

        return_typ = calculate_type_for_external_return(vyper_typ)
        ret = abi.decode_single(return_typ.abi_type.selector_name(), computation.output)

        # unwrap the tuple if needed
        if not isinstance(vyper_typ, TupleType):
            (ret,) = ret

        return vyper_object(ret, vyper_typ)

    def handle_error(self, computation):
        err = computation.error

        # decode error msg if it's "Error(string)"
        # b"\x08\xc3y\xa0" == method_id("Error(string)")
        if isinstance(err.args[0], bytes) and err.args[0][:4] == b"\x08\xc3y\xa0":
            err.args = (
                abi.decode_single("(string)", err.args[0][4:])[0],
                *err.args[1:],
            )

        error_msg = f"{repr(computation.error)} "

        stack_trace = self.vyper_stack_trace(computation)

        for (c, computation) in stack_trace:
            error_msg += f"\n\n{c}\n"

            error_detail = self.find_error_meta(computation.code)
            if error_detail is not None:
                error_msg += f" <dev: {error_detail}>"

            ast_source = c.find_source_of(computation.code)
            if ast_source is not None:
                # VyperException.__str__ does a lot of formatting for us
                error_msg = str(VyperException(error_msg, ast_source))

            frame_detail = c.debug_frame(computation)
            if frame_detail is not None:
                frame_detail.fn_name = "locals"  # override the displayed name
                if len(frame_detail) > 0:
                    error_msg += f" {frame_detail}"

        raise BoaError(error_msg)

    def vyper_stack_trace(self, computation):
        ret = [(self, computation)]
        error_detail = self.find_error_meta(computation.code)
        if error_detail not in EXTERNAL_CALL_ERRORS:
            return ret
        if len(computation.children) < 1:
            return ret
        if not computation.children[-1].is_error:
            return ret
        child = computation.children[-1]
        child_obj = self.env.lookup_contract(child.msg.code_address)
        child_trace = child_obj.vyper_stack_trace(child)
        return child_trace + ret

    # eval vyper code in the context of this contract
    def eval(self, stmt: str) -> Any:
        tmp = self._source_map

        bytecode, source_map, typ = self.compile_stmt(stmt)
        self._source_map = source_map

        method_id = b"dbug"  # note dummy method id, doesn't get validated
        c = self.env.execute_code(
            to_address=self.address, bytecode=bytecode, data=method_id
        )

        ret = self.marshal_to_python(c, typ)

        self._source_map = tmp  # restore

        return ret

    @cached_property
    def _ast_module(self):
        module = copy.deepcopy(self.compiler_data.vyper_module)

        # do the same thing as vyper_module_folded but skip getter expansion
        vy_ast.folding.fold(module)
        with vy_ns.get_namespace().enter_scope():
            validation.add_module_namespace(module, self.compiler_data.interface_codes)
            validation.validate_functions(module)
            # we need to cache the namespace right here(!).
            # set_data_positions will modify the type definitions in place.
            self._cache_namespace(vy_ns.get_namespace())

        vy_ast.expansion.remove_unused_statements(module)
        # calculate slots for all storage variables, tagging
        # the types in the namespace.
        set_data_positions(module, storage_layout_overrides=None)

        return module

    # the global namespace is expensive to compute, so cache it
    def _cache_namespace(self, namespace):
        # copy.copy doesn't really work on Namespace objects, copy by hand
        ret = vy_ns.Namespace()
        ret._scopes = copy.deepcopy(namespace._scopes)
        for s in namespace._scopes:
            for n in s:
                ret[n] = namespace[n]
        self._vyper_namespace = ret

    @contextlib.contextmanager
    def override_vyper_namespace(self):
        # ensure self._vyper_namespace is computed
        m = self._ast_module  # noqa: F841
        try:
            with vy_ns.override_global_namespace(self._vyper_namespace):
                yield
        finally:
            self._vyper_namespace["self"].members.pop("__boa_debug__", None)

    # for eval(), we need unoptimized assembly, since the dead code
    # eliminator might prune a dead function (which we want to eval)
    @cached_property
    def unoptimized_assembly(self):
        runtime = self.compiler_data.ir_runtime
        return compile_ir.compile_to_assembly(runtime, no_optimize=True)

    @cached_property
    def data_section_size(self):
        return self.global_ctx.immutable_section_bytes

    @cached_property
    def data_section(self):
        # extract the data section from the bytecode
        if self.data_section_size:
            return self.bytecode[-self.data_section_size :]
        else:
            return b""

    @cached_property
    def unoptimized_bytecode(self):
        s, _ = compile_ir.assembly_to_evm(
            self.unoptimized_assembly, insert_vyper_signature=True
        )
        return s + self.data_section

    # compile a fragment (single expr or statement) in the context
    # of the contract, returning the ABI encoded value if it is an expr.
    def compile_stmt(self, source_code: str) -> Any:
        # this method is super slow so we cache compilation results
        if source_code in self._eval_cache:
            return self._eval_cache[source_code]

        ast = parse_to_ast(source_code)
        vy_ast.folding.fold(ast)
        ast = ast.body[0]

        typ = None
        return_sig = ""
        debug_body = source_code

        ifaces = self.compiler_data.interface_codes

        if isinstance(ast, vy_ast.Expr):
            with self.override_vyper_namespace():
                try:
                    typ = get_exact_type_from_node(ast.value)
                    return_sig = f"-> {typ}"
                    debug_body = f"return {source_code}"
                except InvalidType:
                    pass

        # wrap code in function so that we can easily generate code for it
        wrapper_code = textwrap.dedent(
            f"""
            @external
            @payable
            def __boa_debug__() {return_sig}:
                {debug_body}
        """
        )

        ast = parse_to_ast(wrapper_code, ifaces)
        vy_ast.folding.fold(ast)

        # annotate ast
        with self.override_vyper_namespace():
            validation.add_module_namespace(ast, self.compiler_data.interface_codes)
            validation.validate_functions(ast)

        ast = ast.body[0]

        sig = FunctionSignature.from_definition(ast, self.global_ctx)
        ast._metadata["signature"] = sig

        sigs = {"self": self.compiler_data.function_signatures}
        ir = generate_ir_for_function(ast, sigs, self.global_ctx, False)
        # generate bytecode where selector check always succeeds
        ir = IRnode.from_list(
            ["with", _METHOD_ID_VAR, abi_method_id(sig.base_signature), ir]
        )

        # skip optimizing; we will optimize a couple lines down anyway
        assembly = compile_ir.compile_to_assembly(ir, no_optimize=True)

        # add original assembly in so that jumpdests in the fragment
        # assemble correctly in final bytecode
        # note this is somewhat kludgy
        assembly.extend(self.unoptimized_assembly)

        # note: optimize_assembly optimizes in place.
        compile_ir._optimize_assembly(assembly)

        bytecode, source_map = compile_ir.assembly_to_evm(assembly)

        # tack on the data section
        bytecode += self.data_section

        # return the bytecode and other artifacts associated with
        # this build
        ret = bytecode, source_map, typ
        self._eval_cache[source_code] = ret
        return ret

    @cached_property
    def _sigs(self):
        sigs = {}
        global_ctx = self.compiler_data.global_ctx
        if global_ctx._contracts or global_ctx._interfaces:
            sigs = parse_external_interfaces(sigs, global_ctx)
        sigs["self"] = self.compiler_data.function_signatures
        return sigs


# inherit from VyperException for pretty tracebacks
class BoaError(VyperException):
    pass


class VyperFunction:
    def __init__(self, fn_ast, contract):
        self.fn_ast = fn_ast
        self.contract = contract
        self.env = contract.env

    def __repr__(self):
        return repr(self.fn_ast)

    @cached_property
    def fn_signature(self):
        return self.contract.compiler_data.function_signatures[self.fn_ast.name]

    @cached_property
    def ir(self):
        # patch compiler_data to have IR for every function
        sigs = self.contract._sigs
        global_ctx = self.contract.global_ctx

        ir = generate_ir_for_function(self.fn_ast, sigs, global_ctx, False)
        return optimize(ir)

    @cached_property
    def assembly(self):
        ir = IRnode.from_list(
            ["with", _METHOD_ID_VAR, ["shr", 224, ["calldataload", 0]], self.ir]
        )
        return compile_ir.compile_to_assembly(ir)

    @cached_property
    def opcodes(self):
        return compiler_output._build_opcodes(self.bytecode)

    @cached_property
    def bytecode(self):
        bytecode, _ = compile_ir.assembly_to_evm(self.assembly)
        return bytecode

    # hotspot, cache the signature computation
    def args_abi_type(self, num_kwargs):
        if not hasattr(self, "_signature_cache"):
            self._signature_cache = {}
        if num_kwargs in self._signature_cache:
            return self._signature_cache[num_kwargs]

        # align the kwargs with the signature
        sig_kwargs = self.fn_signature.default_args[:num_kwargs]
        sig_args = self.fn_signature.base_args + sig_kwargs
        args_abi_type = (
            "(" + ",".join(arg.typ.abi_type.selector_name() for arg in sig_args) + ")"
        )
        method_id = abi_method_id(self.fn_signature.name + args_abi_type).to_bytes(
            4, "big"
        )
        self._signature_cache[num_kwargs] = (method_id, args_abi_type)

        return method_id, args_abi_type

    def _prepare_calldata(self, *args, **kwargs):
        if (
            not len(self.fn_signature.base_args)
            <= len(args)
            <= len(self.fn_signature.args)
        ):
            raise Exception(f"bad args to {self}")

        # align the kwargs with the signature
        # sig_kwargs = self.fn_signature.default_args[: len(kwargs)]

        total_non_base_args = len(kwargs) + len(args) - len(self.fn_signature.base_args)
        method_id, args_abi_type = self.args_abi_type(total_non_base_args)

        # allow things with `.address` to be encode-able
        args = [getattr(arg, "address", arg) for arg in args]

        encoded_args = abi.encode_single(args_abi_type, args)
        return method_id + encoded_args

    def __call__(self, *args, **kwargs):
        calldata_bytes = self._prepare_calldata(*args, **kwargs)
        computation = self.env.execute_code(
            to_address=self.contract.address,
            bytecode=self.contract.bytecode,
            data=calldata_bytes,
        )

        typ = self.fn_signature.return_type
        return self.contract.marshal_to_python(computation, typ)


_typ_cache = {}


def vyper_object(val, vyper_type):
    # make a thin wrapper around whatever type val is,
    # and tag it with _vyper_type metadata

    vt = type(val)
    if vt is bool:
        # https://stackoverflow.com/q/2172189
        # bool is not ambiguous wrt vyper type anyways.
        return val

    if vt not in _typ_cache:
        # ex. class int_wrapper(int): pass
        _typ_cache[vt] = type(f"{vt.__name__}_wrapper", (vt,), {})

    t = _typ_cache[type(val)]

    ret = t(val)
    ret._vyper_type = vyper_type
    return ret



import contextlib
import logging
import sys
from typing import Any, Iterator, Optional, Union

import eth.constants as constants
import eth.tools.builder.chain as chain
import eth.vm.forks.spurious_dragon.computation as spurious_dragon
from eth.chains.mainnet import MainnetChain
from eth.db.atomic import AtomicDB
from eth.vm.code_stream import CodeStream
from eth.vm.message import Message
from eth.vm.opcode_values import STOP
from eth.vm.transaction_context import BaseTransactionContext
from eth_abi import decode_single
from eth_typing import Address
from eth_utils import setup_DEBUG2_logging, to_canonical_address, to_checksum_address


def enable_pyevm_verbose_logging():
    logging.basicConfig()
    logger = logging.getLogger("eth.vm.computation.Computation")
    setup_DEBUG2_logging()
    logger.setLevel("DEBUG2")


class VMPatcher:
    _exc_patchables = {
        # env vars vyper supports
        "block_number": "_block_number",
        "timestamp": "_timestamp",
        "coinbase": "_coinbase",
        "difficulty": "_difficulty",
        "prev_hashes": "_prev_hashes",
        "chain_id": "_chain_id",
    }

    _cmp_patchables = {"code_size_limit": "EIP170_CODE_SIZE_LIMIT"}

    def __init__(self, vm):
        patchables = [
            (self._exc_patchables, vm.state.execution_context),
            (self._cmp_patchables, spurious_dragon),
        ]
        # https://stackoverflow.com/a/12999019
        object.__setattr__(self, "_patchables", patchables)

    def __getattr__(self, attr):
        for s, p in self._patchables:
            if attr in s:
                return getattr(p, s[attr])
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        for s, p in self._patchables:
            if attr in s:
                setattr(p, s[attr], value)
                return

    # to help auto-complete
    def __dir__(self):
        patchable_keys = [k for p, _ in self._patchables for k in p]
        return dir(super()) + patchable_keys

    # save and restore patch values
    @contextlib.contextmanager
    def anchor(self):
        snap = {}
        for s, _ in self._patchables:
            for attr in s:
                snap[attr] = getattr(self, attr)

        try:
            yield

        finally:
            for s, _ in self._patchables:
                for attr in s:
                    setattr(self, attr, snap[attr])


def console_log(computation):
    msgdata = computation.msg.data_as_bytes
    schema, payload = decode_single("(string,bytes)", msgdata[4:])
    data = decode_single(schema, payload)
    print(*data, file=sys.stderr)
    return computation


CONSOLE_ADDRESS = bytes.fromhex("000000000000000000636F6E736F6C652E6C6F67")


AddressT = Union[Address, bytes, str]  # make mypy happy


def _addr(addr: AddressT) -> Address:
    return Address(to_canonical_address(addr))


# a code stream which keeps a trace of opcodes it has executed
class TracingCodeStream(CodeStream):
    __slots__ = [
        "_length_cache",
        "_fake_codesize",
        "_raw_code_bytes",
        "invalid_positions",
        "valid_positions",
        "program_counter",
    ]

    def __init__(self, *args, start_pc=0, fake_codesize=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._trace = []  # trace of opcodes that were run
        self.program_counter = start_pc  # configurable start PC
        self._fake_codesize = fake_codesize  # what CODESIZE returns

    def __iter__(self) -> Iterator[int]:
        # upstream says: "a very performance-sensitive method"
        # note: not clear to me that len(raw_code_bytes) is a hotspot
        while self.program_counter < self._length_cache:
            opcode = self._raw_code_bytes[self.program_counter]

            self._trace.append(self.program_counter)
            self.program_counter += 1
            yield opcode

        yield STOP

    def __len__(self):
        if self._fake_codesize is not None:
            return self._fake_codesize
        return self._length_cache


# ### section: sha3 preimage tracing
# (TODO: move to dedicated module)
def to_int(value):
    if isinstance(value, tuple):
        return to_int(value[1])  # how py-evm stores stuff on stack
    if isinstance(value, int):
        return value
    if isinstance(value, bytes):
        return int.from_bytes(value, "big")

    raise ValueError("invalid type %s", type(value))


def to_bytes(value):
    if isinstance(value, tuple):
        return to_bytes(value[1])  # how py-evm stores stuff on stack
    if isinstance(value, bytes):
        return value
    if isinstance(value, int):
        return value.to_bytes(32, "big")

    raise ValueError("invalid type %s", type(value))


class Sha3PreimageTracer:
    mnemonic = "SHA3"

    # trace preimages of sha3

    def __init__(self, sha3_op, preimage_map):
        self.preimages = preimage_map
        self.sha3 = sha3_op

    def __call__(self, computation):
        size, offset = [to_int(x) for x in computation._stack.values[-2:]]

        # dispatch into py-evm
        self.sha3(computation)

        if size != 64:
            return

        preimage = computation._memory.read_bytes(offset, size)

        image = to_bytes(computation._stack.values[-1])

        self.preimages[image] = preimage


class SstoreTracer:
    mnemonic = "SSTORE"

    def __init__(self, sstore_op, trace_db):
        self.trace_db = trace_db
        self.sstore = sstore_op

    def __call__(self, computation):
        value, slot = [to_bytes(t) for t in computation._stack.values[-2:]]
        account = to_checksum_address(computation.msg.to)

        self.trace_db.setdefault(account, set())
        # we don't want to deal with snapshots/commits/reverts, so just
        # register that the slot was touched and downstream can filter
        # zero entries.
        self.trace_db[account].add(slot)

        # dispatch into py-evm
        self.sstore(computation)


# ### End section: sha3 tracing


class TrivialGasMeter:
    def __init__(self, start_gas):
        self.start_gas = start_gas
        self.gas_remaining = start_gas

    def consume_gas(self, amount, reason):
        pass

    def refund_gas(self, amount):
        pass

    def return_gas(self, amount):
        pass


# wrapper class around py-evm which provides a "contract-centric" API
class Env:
    _singleton = None
    _initial_address_counter = 100

    def __init__(self):
        self.chain = _make_chain()
        self.vm = self.chain.get_vm()
        self._gas_price = 0

        self._address_counter = self.__class__._initial_address_counter

        self._aliases = {}

        # TODO differentiate between origin and sender
        self.eoa = self.generate_address("root")

        class OpcodeTracingComputation(self.vm.state.computation_class):
            _gas_metering = True

            def __init__(self, *args, **kwargs):
                # super() hardcodes CodeStream into the ctor
                # so we have to override it here
                super().__init__(*args, **kwargs)
                self.code = TracingCodeStream(
                    self.code._raw_code_bytes,
                    fake_codesize=getattr(self.msg, "_fake_codesize", None),
                    start_pc=getattr(self.msg, "_start_pc", 0),
                )
                if not self.__class__._gas_metering:
                    self._gas_meter = TrivialGasMeter(self.msg.gas)

                self._precompiles[CONSOLE_ADDRESS] = console_log

        # TODO make metering toggle-able
        c = OpcodeTracingComputation

        self.vm.state.computation_class = c

        # patch in tracing opcodes
        self.sha3_trace = {}
        self.sstore_trace = {}
        c.opcodes[0x20] = Sha3PreimageTracer(c.opcodes[0x20], self.sha3_trace)
        c.opcodes[0x55] = SstoreTracer(c.opcodes[0x55], self.sstore_trace)

        self.vm.patch = VMPatcher(self.vm)

        self._contracts = {}

    def set_gas_metering(self, val: bool) -> None:
        self.vm.state.computation_class._gas_metering = val

    def register_contract(self, address, obj):
        self._contracts[to_checksum_address(address)] = obj

    def lookup_contract(self, address):
        return self._contracts.get(to_checksum_address(address))

    def alias(self, address, name):
        self._aliases[to_checksum_address(address)] = name

    def lookup_alias(self, address):
        return self._aliases[to_checksum_address(address)]

    # context manager which snapshots the state and reverts
    # to the snapshot on exiting the with statement
    @contextlib.contextmanager
    def anchor(self):
        snapshot_id = self.vm.state.snapshot()
        try:
            with self.vm.patch.anchor():
                yield
        finally:
            self.vm.state.revert(snapshot_id)

    # TODO is this a good name
    @contextlib.contextmanager
    def prank(self, address):
        tmp = self.eoa
        self.eoa = to_checksum_address(address)
        try:
            yield
        finally:
            self.eoa = tmp

    @classmethod
    def get_singleton(cls):
        if cls._singleton is None:
            cls._singleton = cls()
        return cls._singleton

    def generate_address(self, alias: Optional[str] = None) -> AddressT:
        self._address_counter += 1
        t = self._address_counter.to_bytes(length=20, byteorder="big")
        # checksum addr easier for humans to debug
        ret = to_checksum_address(t)
        if alias is not None:
            self.alias(ret, alias)

        return ret


    def deploy_code(
        self,
        deploy_to: AddressT = constants.ZERO_ADDRESS,
        sender: Optional[AddressT] = None,
        gas: int = None,
        value: int = 0,
        bytecode: bytes = b"",
        data: bytes = b"",
        start_pc: int = 0,
    ) -> bytes:
        if gas is None:
            gas = self.vm.state.gas_limit
        if sender is None:
            sender = self.eoa

        msg = Message(
            to=_addr(deploy_to),
            sender=_addr(sender),
            gas=gas,
            value=value,
            code=bytecode,
            data=data,
        )
        tx_ctx = BaseTransactionContext(origin=_addr(sender), gas_price=self._gas_price)
        c = self.vm.state.computation_class.apply_create_message(
            self.vm.state, msg, tx_ctx
        )

        if c.is_error:
            raise c.error
        return c.output

    def execute_code(
        self,
        to_address: AddressT = constants.ZERO_ADDRESS,
        sender: AddressT = None,
        gas: int = None,
        value: int = 0,
        bytecode: bytes = b"",
        data: bytes = b"",
        start_pc: int = 0,
        fake_codesize: int = None,
    ) -> Any:
        if gas is None:
            gas = self.vm.state.gas_limit
        if sender is None:
            sender = self.eoa

        class FakeMessage(Message):  # Message object with settable attrs
            __dict__: dict = {}

        msg = FakeMessage(
            sender=_addr(sender),
            to=_addr(to_address),
            gas=gas,
            value=value,
            code=bytecode,
            data=data,
        )
        msg._fake_codesize = fake_codesize  # type: ignore
        msg._start_pc = start_pc  # type: ignore
        tx_ctx = BaseTransactionContext(origin=_addr(sender), gas_price=self._gas_price)
        return self.vm.state.computation_class.apply_message(self.vm.state, msg, tx_ctx)


GENESIS_PARAMS = {"difficulty": constants.GENESIS_DIFFICULTY, "gas_limit": int(1e8)}


# TODO make fork configurable - ex. "latest", "frontier", "berlin"
# TODO make genesis params+state configurable
def _make_chain():
    # TODO should we use MiningChain? is there a perf difference?
    # TODO debug why `fork_at()` cannot accept 0 as block num
    _Chain = chain.build(MainnetChain, chain.latest_mainnet_at(1))
    return _Chain.from_genesis(AtomicDB(), GENESIS_PARAMS)
